@@ -151,20 +151,17 @@ grd_session_vnc_queue_resize_framebuffer (GrdSessionVnc *session_vnc,
 }
 
 void
-grd_session_vnc_draw_buffer (GrdSessionVnc *session_vnc,
+grd_session_vnc_take_buffer (GrdSessionVnc *session_vnc,
                              void          *data)
 {
-  size_t size;
-
   if (session_vnc->pending_framebuffer_resize)
-    return;
+    {
+      free (data);
+      return;
+    }
 
-  size = (session_vnc->rfb_screen->height *
-          grd_session_vnc_get_framebuffer_stride (session_vnc));
-
-  memcpy (session_vnc->rfb_screen->frameBuffer,
-          data,
-          size);
+  free (session_vnc->rfb_screen->frameBuffer);
+  session_vnc->rfb_screen->frameBuffer = data;
 
   rfbMarkRectAsModified (session_vnc->rfb_screen, 0, 0,
                          session_vnc->rfb_screen->width,
@@ -172,6 +169,12 @@ grd_session_vnc_draw_buffer (GrdSessionVnc *session_vnc,
   rfbProcessEvents (session_vnc->rfb_screen, 0);
 }
 
+void
+grd_session_vnc_flush (GrdSessionVnc *session_vnc)
+{
+  rfbProcessEvents (session_vnc->rfb_screen, 0);
+}
+
 void
 grd_session_vnc_set_cursor (GrdSessionVnc *session_vnc,
                             rfbCursorPtr   rfb_cursor)
@@ -675,8 +678,8 @@ close_session_idle (gpointer user_data)
 }
 
 static void
-on_pipwire_stream_closed (GrdVncPipeWireStream *stream,
-                          GrdSessionVnc        *session_vnc)
+on_pipewire_stream_closed (GrdVncPipeWireStream *stream,
+                           GrdSessionVnc        *session_vnc)
 {
   g_warning ("PipeWire stream closed, closing client");
 
@@ -702,7 +705,7 @@ grd_session_vnc_stream_ready (GrdSession *session,
     }
 
   g_signal_connect (session_vnc->pipewire_stream, "closed",
-                    G_CALLBACK (on_pipwire_stream_closed),
+                    G_CALLBACK (on_pipewire_stream_closed),
                     session_vnc);
 
   if (!session_vnc->source)
