@@ -24,7 +24,10 @@
 #define GRD_TYPES_H
 
 typedef struct _GrdContext GrdContext;
+typedef struct _GrdRdpSAMFile GrdRdpSAMFile;
+typedef struct _GrdRdpServer GrdRdpServer;
 typedef struct _GrdSession GrdSession;
+typedef struct _GrdSessionRdp GrdSessionRdp;
 typedef struct _GrdSessionVnc GrdSessionVnc;
 typedef struct _GrdStream GrdStream;
 typedef struct _GrdPipeWireStream GrdPipeWireStream;
